from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bunkr.api.upload import Chunk, FileUpload


class BunkrUploaderError(Exception): ...


class ChunkUploadError(BunkrUploaderError):
    def __init__(self, file: FileUpload, chunk: Chunk) -> None:
        self.message = f"Failed uploading chunk #{chunk.index + 1}/{chunk.total} of {file.uuid} ('{file.name}')"
        super().__init__(self.message)


class FileUploadError(BunkrUploaderError):
    def __init__(self, file: FileUpload) -> None:
        self.message = f"Failed to upload {file.path}"
        super().__init__(self.message)



from __future__ import annotations

import asyncio
import dataclasses
import datetime  # noqa: TC003
import logging
from typing import TYPE_CHECKING, ClassVar, Self

from pydantic import TypeAdapter

from bunkr import aio, progress
from bunkr.api import ApiProxy, BunkrAPI
from bunkr.api.errors import ChunkUploadError, FileUploadError
from bunkr.api.responses import FileResponse, UploadResponse
from bunkr.api.upload import Chunk, FileUpload
from bunkr.logger import utc_now

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable
    from pathlib import Path

    from yarl import URL

    from bunkr.config import Config


logger = logging.getLogger(__name__)


@dataclasses.dataclass(slots=True)
class FileUploadResult:
    file: FileUpload
    result: UploadResponse
    timestamp: datetime.datetime = dataclasses.field(init=False, default_factory=utc_now)

    _serializer: ClassVar[Callable[[Self], bytes] | None] = None

    @classmethod
    def __new__(cls) -> Self:
        if cls._serializer is None:
            cls._serializer = TypeAdapter(cls).dump_json
        return super(FileUploadResult, cls).__new__(cls)

    def __str__(self) -> str:
        assert self._serializer is not None
        return self._serializer(self).decode()


@dataclasses.dataclass(slots=True)
class BunkrUploader(ApiProxy):
    config: Config
    upload_callback: Callable[[FileUploadResult], None] = lambda _: None

    _api: BunkrAPI = dataclasses.field(init=False)
    _sem: asyncio.BoundedSemaphore = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        self._api = BunkrAPI(self.config.token, self.config.chunk_size or 0)
        self._sem = asyncio.BoundedSemaphore(self.config.concurrent_uploads)

    async def _direct_upload(self, upload: FileUpload, server: URL) -> UploadResponse:
        with progress.new_upload(upload.name, upload.size) as hook:
            result = await self._api.upload(upload, server)
            hook.advance(upload.size)
            return result

    async def _chunked_upload(self, upload: FileUpload, server: URL) -> UploadResponse:
        with progress.new_upload(upload.name, upload.size) as hook:
            chunk_gen = _iter_chunked(upload, self._api.chunk_size)
            try:
                async for chunk in chunk_gen:
                    _ = await self.__upload_chunk(upload, server, chunk)
                    hook.advance(len(chunk.data))
            finally:
                await chunk_gen.aclose()

            return await self._api.finish_chunks(upload, server)

    async def __upload_chunk(self, upload: FileUpload, server: URL, chunk: Chunk) -> None:
        """Upload a single chunk with retry mechanism."""
        for attempt in range(self.config.chunk_retries):
            msg = (
                f"Uploading chunk {chunk.index + 1} of file '{upload.name}'"
                f" (attempt {attempt + 1}/{self.config.chunk_retries})"
            )
            logger.debug(msg)
            try:
                await self._api.upload_chunk(upload, server, chunk)

            except ChunkUploadError as e:
                if attempt < self.config.chunk_retries - 1:
                    logger.error(str(e))
                    await asyncio.sleep(self.config.delay)
                    continue
                raise FileUploadError(upload) from e.__cause__

    async def _upload(self, upload: FileUpload, server: URL) -> UploadResponse:
        """Upload a file in chunks with retry mechanism."""
        info = await self._api.check()
        for attempt in range(self.config.retries):
            try:
                if upload.size <= info.chunkSize.max:
                    return await self._direct_upload(upload, server)

                else:
                    return await self._chunked_upload(upload, server)

            except FileUploadError as e:
                cause = e.__cause__ or e
                if attempt < self.config.retries - 1:
                    msg = f"{cause} (attempt {attempt + 1}/{self.config.retries})"
                    logger.error(msg)
                    await asyncio.sleep(self.config.delay)
                    continue

                msg = f"Skipping upload of '{upload.path}' after {self.config.retries} failed attempt(s) ({str(cause)[:40]}"
                logger.error(msg, exc_info=cause)

        return UploadResponse(success=False, files=[FileResponse(name=upload.uuid, url=None)])

    async def _request_upload_server(self) -> URL:
        node_response = await self._api.node()
        if not node_response.success:
            raise RuntimeError("Unable to get server to upload")

        return node_response.url.with_path("/api/")

    async def _get_album_id(self, album_name: str) -> int:
        album_name_ci = album_name.casefold()
        iter_albums = self._api.iter_albums()
        try:
            async for albums in iter_albums:
                for album in albums:
                    if album.name.casefold() == album_name_ci:
                        return album.id
        finally:
            await iter_albums.aclose()

        logger.info(f"Album '{album_name}' does not exists, creating")
        album = await self._api.create_album(album_name, description=album_name)
        return album.id

    async def _prepare_uploads(self, path: Path, *, recurse: bool) -> tuple[FileUpload, ...]:
        files = await asyncio.to_thread(_get_files, path, recurse=recurse)
        info = await self._api.check()
        human_max_size = info.maxSize.human_readable(decimal=True)

        async def _prepare_upload(path: Path) -> FileUpload | None:
            try:
                if path.suffix.casefold() in info.stripTags.blacklistExtensions:
                    logger.error(f"File '{path}' has blacklisted extension: {path.suffix}")
                    return

                size = (await asyncio.to_thread(path.stat)).st_size

                if size > info.maxSize:
                    msg = f"File '{path}' ({size:,} B) is bigger than max file size: {info.maxSize:,} B ({human_max_size})"
                    logger.error(msg)
                    return

                upload = await asyncio.to_thread(FileUpload.create, path, size)
            except Exception:
                logger.exception(f"Unable to prepare upload of '{path}'")
                return

            else:
                return upload

        return tuple(filter(None, await asyncio.gather(*map(_prepare_upload, files))))

    async def upload(
        self, path: Path, *, recurse: bool = False, album: str | None = None
    ) -> list[FileUploadResult]:
        files_to_upload = await self._prepare_uploads(path, recurse=recurse)
        if not files_to_upload:
            logger.error("No files left to upload")
            return []

        album_id = None
        album = album or self.config.album
        if album:
            album_id = str(await self._get_album_id(album))
            logger.debug(f"album id: '{album_id}'")

        tasks: list[asyncio.Task[FileUploadResult | None]] = []

        with progress.new_progress():
            async with asyncio.TaskGroup() as tg:
                for file in files_to_upload:
                    _ = await self._sem.acquire()
                    tasks.append(
                        tg.create_task(self._try_upload(file, album_id)),
                    )

        return [result for t in tasks if (result := t.result()) if not None]

    async def _try_upload(
        self, upload: FileUpload, album_id: str | None
    ) -> FileUploadResult | None:
        upload.album_id = album_id
        try:
            server = await self._request_upload_server()
            logger.info(f"Using {server = !s} for upload of '{upload.path}'")
            response = await self._upload(upload, server)
        except Exception:
            logger.exception(f"Upload of '{upload.path}' failed")
        else:
            result = FileUploadResult(upload, response)
            self.upload_callback(result)
            return result
        finally:
            self._sem.release()


def _get_files(path: Path, *, recurse: bool) -> list[Path]:
    if not path.exists():
        raise FileNotFoundError(path)
    if path.is_file():
        files_to_upload = [path]

    elif recurse:
        files_to_upload = path.rglob("*")
    else:
        files_to_upload = path.iterdir()

    return sorted(
        (f for f in files_to_upload if f.is_file()), key=lambda p: p.as_posix().casefold()
    )


async def _iter_chunked(upload: FileUpload, chunk_size: int) -> AsyncGenerator[Chunk]:
    """Iterate over file chunks."""
    n_chunks = (upload.size + chunk_size - 1) // chunk_size
    index = 0
    async with aio.open(upload.path, mode="rb") as fp:
        while data := await fp.read(chunk_size):
            offset = chunk_size * index
            yield Chunk(data, index, n_chunks, offset)
            index += 1



from typing import Annotated

from cyclopts import Parameter
from pydantic import BaseModel, ByteSize, Field


class Config(BaseModel, defer_build=True):
    token: Annotated[str, Parameter(alias="-t", env_var="BUNKR_TOKEN")]
    "API token for your account so that you can upload to a specific account/folder. You can also set the BUNKR_TOKEN environment variable for this"

    album: Annotated[str | None, Parameter(alias="-n")] = None
    "Name to use for album. If an album with this name already exists, add the files to that album"

    concurrent_uploads: Annotated[int, Parameter(alias="-c")] = Field(2, gt=0, le=50)
    "Maximum parallel uploads to do at once"

    chunk_size: ByteSize | None = Field(default=None, gt=0)
    "Size of chunks to use for uploads. 0 or `None` will use the server's maximum chunk size"

    public: bool = True
    "Make all uploaded files public"

    retries: Annotated[int, Parameter(alias="-R")] = Field(default=1, gt=0)
    "How many times to retry a failed file upload"

    chunk_retries: int = Field(default=2, gt=0)
    "How many times to retry a failed chunk upload"

    delay: float = Field(default=1.0, ge=0)
    "How many seconds to wait in between failed upload attempts"



from __future__ import annotations

import dataclasses
import datetime
import json
import re
from html import unescape
from pathlib import Path
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Self

from yarl import URL

if TYPE_CHECKING:
    from collections.abc import Generator


@dataclasses.dataclass(slots=True)
class File:
    id: int
    name: str
    original: str
    slug: str
    type: str
    extension: str
    size: int
    timestamp: datetime.datetime
    thumbnail: str
    cdnEndpoint: str  # noqa: N815

    src: URL | None = None

    def __post_init__(self) -> None:
        if self.thumbnail.count("https://") != 1:
            return

        if URL(self.thumbnail, encoded="%" in self.thumbnail).parts[1:2] != ("thumbs",):
            return

        src_str = self.thumbnail.replace("/thumbs/", "/")
        src = (
            URL(src_str, encoded="%" in src_str)
            .with_suffix(Path(self.name).suffix)
            .with_query(None)
        )
        if src.suffix.lower() not in _IMAGE_EXTS:
            assert src.host
            src = src.with_host(src.host.replace("i-", ""))
        self.src = src

    def __json__(self) -> dict[str, Any]:
        me = dataclasses.asdict(self)
        me["timestamp"] = str(self.timestamp)
        if self.src:
            me["src"] = str(self.src)
        return me


_translation_map = MappingProxyType(
    {f" {field.name}: ": f'"{field.name}": ' for field in dataclasses.fields(File)}
)
_escape_attrs = re.compile("|".join(sorted(_translation_map.keys(), key=len, reverse=True))).sub


def _fix_unicode(value: Any) -> Any:
    if isinstance(value, str):
        return value.encode("raw_unicode_escape").decode("unicode-escape")
    return value


def _decode_files(text: str) -> Generator[File]:
    content = _escape_attrs(lambda m: _translation_map[m.group(0)], text.replace("\\'", "'"))

    file: dict[str, Any]
    for file in json.loads(content):
        file = {k: _fix_unicode(v) for k, v in file.items()}
        timestamp = datetime.datetime.strptime(file.pop("timestamp"), "%H:%M:%S %d/%m/%Y").replace(
            tzinfo=datetime.UTC
        )
        yield File(timestamp=timestamp, **file)


@dataclasses.dataclass(slots=True, order=True)
class PublicAlbum:
    id: int
    slug: str
    name: str
    files: tuple[File, ...]

    @classmethod
    def parse(cls, slug: str, html: str) -> Self:
        def extr(before: str, after: str) -> str:
            start = html.index(before) + len(before)
            end = html.index(after, start)
            return html[start:end]

        id_ = int(extr('albumID = "', '";'))
        name = unescape(extr('<meta property="og:title" content="', '">'))

        album_js = extr("window.albumFiles = ", "</script>")
        files = _decode_files(album_js[: album_js.rindex("];") + 1])
        return cls(id_, slug, name, tuple(files))

    def __json__(self) -> dict[str, Any]:
        return dict(  # noqa: C408
            id=self.id,
            slug=self.slug,
            name=self.name,
            files=tuple(file.__json__() for file in self.files),
        )

    def __str__(self) -> str:
        import json

        return json.dumps(self.__json__(), indent=2, ensure_ascii=False)


_IMAGE_EXTS = frozenset(
    {
        ".gif",
        ".gifv",
        ".heic",
        ".jfif",
        ".jif",
        ".jpe",
        ".jpeg",
        ".jpg",
        ".jxl",
        ".png",
        ".svg",
        ".tif",
        ".tiff",
        ".webp",
    }
)



import asyncio
import dataclasses
import itertools
import logging
from collections.abc import AsyncGenerator, Mapping
from json import dumps as json_dumps
from typing import Any, Self

from aiohttp import ClientSession, ClientTimeout, FormData
from multidict import CIMultiDict
from pydantic.type_adapter import TypeAdapter
from yarl import URL

from bunkr.api.album import PublicAlbum
from bunkr.api.errors import ChunkUploadError, FileUploadError
from bunkr.api.responses import (
    Album,
    CreateAlbumResponse,
    InfoResponse,
    NodeResponse,
    UploadResponse,
    VerifyTokenResponse,
)
from bunkr.api.upload import Chunk, FileUpload

_API_ENTRYPOINT = URL("https://dash.bunkr.cr/api/")
_SEMAPHORE = asyncio.Semaphore(50)
_parse_albums = TypeAdapter(list[Album]).validate_python

logger = logging.getLogger(__name__)


@dataclasses.dataclass(slots=True)
class BunkrAPI:
    _token: str | None = None
    _chunk_size: int = 0
    _session: ClientSession | None = dataclasses.field(init=False, default=None)
    _info: InfoResponse | None = dataclasses.field(init=False, default=None)

    @property
    def chunk_size(self) -> int:
        return self._chunk_size

    async def connect(self) -> None:
        info = await self.check()
        self._chunk_size = self.chunk_size or info.chunkSize.max
        if self._chunk_size > info.chunkSize.max:
            msg = f"Chunk size is too high. Using max chunksize ({info.chunkSize.max.human_readable(decimal=True)})"
            logger.warning(msg)
            self._chunk_size = info.chunkSize.max

        if self._token:
            _ = await self.verify_token(self._token)

    async def __aenter__(self) -> Self:
        await self.connect()
        return self

    async def __aexit__(self, *_: Any) -> None:
        await self.aclose()

    async def aclose(self) -> None:
        if self._session is not None:
            await self._session.close()

    @property
    def session(self) -> ClientSession:
        if self._session is None:
            self._session = ClientSession(
                headers={
                    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:142.0) Gecko/20100101 Firefox/142.0",
                    "Referer": "https://dash.bunkr.cr/",
                    "striptags": "null",
                    "Origin": "https://dash.bunkr.cr",
                },
                raise_for_status=True,
                timeout=ClientTimeout(sock_connect=30, sock_read=20),
            )
        return self._session

    async def _request_json(
        self,
        path_or_url: URL | str,
        *,
        form: FormData | None = None,
        headers: Mapping[str, str] | None = None,
        **json: Any,
    ) -> dict[str, Any]:
        method = "POST" if (form or json) else "GET"
        url = _API_ENTRYPOINT / path_or_url if isinstance(path_or_url, str) else path_or_url

        async with (
            _SEMAPHORE,
            self.session.request(
                method,
                url,
                data=form,
                json=json or None,
                headers=self._prepare_json_headers(headers),
            ) as resp,
        ):
            data = await resp.json()
            record = {
                "url": resp.url,
                "headers": dict(resp.headers),
                "data": data,
            }
            logger.debug(json_dumps(record, indent=4, default=str))

            return data

    def _prepare_json_headers(self, headers: Mapping[str, str] | None = None) -> CIMultiDict[str]:
        """Add default headers and transform it to CIMultiDict"""

        combined = CIMultiDict(Accept="application/json, text/plain")
        if self._token:
            combined["token"] = self._token

        if headers:
            headers = CIMultiDict(headers)
            new: set[str] = set()
            for key, value in headers.items():
                if key in new:
                    combined.add(key, value)
                else:
                    combined[key] = value
                    new.add(key)
        return combined

    async def check(self) -> InfoResponse:
        if not self._info:
            response = await self._request_json("check")
            self._info = InfoResponse.model_validate(response)
        return self._info

    async def node(self) -> NodeResponse:
        response = await self._request_json("node")
        return NodeResponse.model_validate(response)

    async def verify_token(self, token: str) -> VerifyTokenResponse:
        response = await self._request_json("tokens/verify", token=token)
        try:
            resp = VerifyTokenResponse.model_validate(response)
        except ValueError:
            raise ValueError("Invalid Token") from None
        if not resp.success:
            raise ValueError("Invalid Token")
        return resp

    async def albums(self) -> list[Album]:
        albums: list[Album] = []
        async for new_albums in self.iter_albums():
            albums.extend(new_albums)
        return albums

    async def iter_albums(self) -> AsyncGenerator[list[Album]]:
        for page in itertools.count(0):
            response = await self._request_json(f"albums/{page}")
            new_albums = _parse_albums(response["albums"])
            yield new_albums
            if len(new_albums) < 50:
                break

    async def public_album(self, url_or_slug: URL | str, /) -> PublicAlbum:
        url = (
            URL(f"https://bunkr.cr/a/{url_or_slug}", encoded="%" in url_or_slug)
            if isinstance(url_or_slug, str) and "/" not in url_or_slug
            else URL(url_or_slug)
        )
        async with self.session.get(url.with_query(advanced=1)) as resp:
            page = await resp.text()

        return PublicAlbum.parse(url.name, page)

    async def create_album(
        self,
        name: str,
        *,
        description: str = "",
        public: bool = True,
        download: bool = True,
    ) -> CreateAlbumResponse:
        response = await self._request_json(
            "albums",
            name=name,
            description=description,
            public=public,
            download=download,
        )
        return CreateAlbumResponse.model_validate(response)

    async def upload(
        self,
        upload: FileUpload,
        server: URL,
        album_id: str | None = None,
    ) -> UploadResponse:
        upload.album_id = upload.album_id or album_id
        try:
            form = FormData()
            form.add_field(
                "files[]",
                value=await asyncio.to_thread(upload.path.read_bytes),
                filename=upload.upload_name,
                content_type=upload.mimetype,
            )
            headers = {"albumid": upload.album_id} if upload.album_id else None
            response = await self._request_json(server / "upload", form=form, headers=headers)

        except Exception as e:
            raise FileUploadError(upload) from e

        result = UploadResponse.model_validate(response)
        if not result.success:
            raise FileUploadError(upload)
        return result

    async def upload_chunk(self, file: FileUpload, server: URL, chunk: Chunk) -> None:
        try:
            form = _create_chunk_form(file, chunk, self.chunk_size)
            result = await self._request_json(server / "upload", form=form)
        except Exception as e:
            raise ChunkUploadError(file, chunk) from e

        if not result["success"]:
            raise ChunkUploadError(file, chunk)

    async def finish_chunks(
        self, file: FileUpload, server: URL, album_id: str | None = None
    ) -> UploadResponse:
        file.album_id = album_id = file.album_id or album_id

        for retry in (True, False):
            try:
                response = await self._request_json(
                    server / "upload/finishchunks",
                    files=[
                        {
                            "uuid": file.uuid,
                            "original": file.name,
                            "type": file.mimetype,
                            "albumid": album_id or None,
                            "filelength": None,
                            "age": None,
                        }
                    ],
                )
                break
            except Exception as e:
                if retry:
                    continue
                raise FileUploadError(file) from e

        else:
            raise FileUploadError(file)

        result = UploadResponse.model_validate(response)
        if not result.success:
            raise FileUploadError(file)
        return result


class ApiProxy:
    _api: BunkrAPI  # pyright: ignore[reportUninitializedInstanceVariable]

    async def __aenter__(self) -> Self:
        _ = await self._api.__aenter__()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self._api.__aexit__(*args)


def _create_chunk_form(file: FileUpload, chunk: Chunk, chunk_size: int) -> FormData:
    form = FormData()
    form.add_fields(
        ("dzuuid", file.uuid),
        ("dzchunkindex", str(chunk.index)),
        ("dztotalfilesize", str(file.size)),
        ("dzchunksize", str(chunk_size)),
        ("dztotalchunkcount", str(chunk.total)),
        ("dzchunkbyteoffset", str(chunk.offset)),
    )
    form.add_field(
        "files[]",
        chunk.data,
        filename=file.upload_name,
        content_type="application/octet-stream",
    )
    return form



from __future__ import annotations

import dataclasses
import mimetypes
from pathlib import Path  # noqa: TC003
from uuid import uuid4

_MAX_FILENAME_LENGTH: int = 240


def _truncate_name(path: Path) -> str:
    if len(path.name) <= _MAX_FILENAME_LENGTH:
        return path.name

    max_bytes = _MAX_FILENAME_LENGTH - len(path.suffix) - 2
    new_stem = path.name.encode("utf-8")[:max_bytes].decode("utf-8", "ignore")
    return f"{new_stem}..{path.suffix}"


@dataclasses.dataclass(slots=True, eq=False)
class Chunk:
    data: bytes
    index: int
    total: int
    offset: int


@dataclasses.dataclass(slots=True, kw_only=True)
class FileUpload:
    path: Path
    name: str
    upload_name: str
    size: int
    mimetype: str
    uuid: str
    album_id: str | None = None

    @staticmethod
    def create(path: Path, /, size: int | None = None) -> FileUpload:
        if size is None:
            size = path.stat().st_size

        return FileUpload(
            path=path,
            name=path.name,
            upload_name=_truncate_name(path),
            size=size,
            mimetype=mimetypes.guess_type(path)[0] or "application/octet-stream",
            uuid=str(uuid4()),
        )



import contextlib
import dataclasses
from collections.abc import Callable, Generator
from contextvars import ContextVar
from typing import Self

from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    SpinnerColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from bunkr.logger import CONSOLE

_progress: ContextVar[Progress] = ContextVar("_progress")


@contextlib.contextmanager
def new_progress() -> Generator[None]:
    token = _progress.set(
        Progress(
            SpinnerColumn(),
            "[progress.description]{task.description}",
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>6.2f}%",
            "━",
            DownloadColumn(),
            "━",
            TransferSpeedColumn(),
            "━",
            TimeRemainingColumn(),
            console=CONSOLE,
            expand=True,
            transient=True,
        )
    )
    try:
        yield
    finally:
        _progress.reset(token)


@dataclasses.dataclass(slots=True)
class ProgressHook:
    advance: Callable[[int], None]
    done: Callable[[], None]

    _done: bool = dataclasses.field(init=False, default=False)

    def __enter__(self) -> Self:
        if self._done:
            raise RuntimeError
        return self

    def __exit__(self, *_) -> None:
        if self._done:
            raise RuntimeError
        self.done()
        self._done = True


def new_upload(file_name: str, file_size: int) -> ProgressHook:
    progress = _progress.get()
    task_id = progress.add_task(file_name, total=file_size)

    def advance(amount: int = 1) -> None:
        progress.advance(task_id, amount)

    def on_exit() -> None:
        progress.remove_task(task_id)

    return ProgressHook(advance, on_exit)



# ruff: noqa: N815
import dataclasses
from datetime import datetime, timedelta
from typing import Annotated

import yarl
from pydantic import BaseModel, ByteSize, HttpUrl, PlainValidator

HttpURL = Annotated[yarl.URL, PlainValidator(lambda url: yarl.URL(str(HttpUrl(url))))]


@dataclasses.dataclass(slots=True)
class ChunkSize:
    max: ByteSize
    default: ByteSize
    timeout: timedelta


@dataclasses.dataclass(slots=True)
class FileIDLimits:
    min: int
    max: int
    default: int
    force: bool


@dataclasses.dataclass(slots=True)
class StripTags:
    blacklistExtensions: set[str]
    default: bool
    force: bool
    video: bool


@dataclasses.dataclass(slots=True)
class Permissions:
    admin: bool
    moderator: bool
    superadmin: bool
    user: bool
    vip: bool
    vvip: bool


@dataclasses.dataclass(slots=True)
class Album:
    descriptionHtml: str
    download: bool
    editedAt: datetime
    enabled: bool
    id: int
    identifier: str
    name: str
    public: bool
    size: ByteSize
    timestamp: datetime
    uploads: int
    zipGeneratedAt: datetime
    zipSize: ByteSize | None


class _Response(BaseModel, populate_by_name=True, defer_build=True):
    description: str = ""
    success: bool = True


class FileResponse(_Response):
    name: str
    url: HttpURL | None

    def model_post_init(self, *_) -> None:
        if self.url is None:
            self.success = False


class UploadResponse(_Response):
    files: list[FileResponse] = []


class CreateAlbumResponse(_Response):
    id: int


class VerifyTokenResponse(_Response):
    defaultRetentionPeriod: timedelta
    group: str
    permissions: Permissions
    retentionPeriods: list[timedelta]
    username: str


class InfoResponse(_Response):
    chunkSize: ChunkSize
    defaultTemporaryUploadAge: int
    enableUserAccounts: bool
    fileIdentifierLength: FileIDLimits
    maintenance: bool
    maxSize: ByteSize
    private: bool
    stripTags: StripTags
    temporaryUploadAges: list[int]


class NodeResponse(_Response):
    url: HttpURL



from __future__ import annotations

import asyncio
import contextlib
import datetime
import logging
from pathlib import Path
from typing import TYPE_CHECKING

from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable


CONSOLE = Console(
    theme=Theme(
        {
            "logging.level.warning": "yellow",
            "logging.level.debug": "blue",
            "logging.level.info": "white",
            "logging.level.error": "red",
        }
    )
)
_ROTATE: bool = True


def utc_now() -> datetime.datetime:
    return datetime.datetime.now(datetime.UTC).replace(microsecond=0)


class NoTracebackFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        record.message = record.getMessage()
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        return self.formatMessage(record)


@contextlib.asynccontextmanager
async def setup_logger() -> AsyncGenerator[Callable[[object], None]]:
    logger = logging.getLogger("bunkr")
    logger.setLevel(logging.DEBUG)
    console_handler = RichHandler(
        show_time=False,
        rich_tracebacks=False,
        tracebacks_show_locals=False,
        show_path=False,
        level=logging.INFO,
        console=CONSOLE,
    )
    console_handler.setFormatter(NoTracebackFormatter())
    logger.addHandler(console_handler)

    log_folder = Path.cwd() / "bunkr_uploader_logs"
    log_folder.mkdir(exist_ok=True)

    if _ROTATE:
        now = utc_now().replace(tzinfo=None).isoformat().replace(":", "").replace(" ", "_")
        name = f"{now}.log"
    else:
        name = "bunkr.log"

    log_file_path = log_folder / name
    jsonl_path = log_file_path.with_suffix(".results.jsonl")
    with log_file_path.open("w", encoding="utf8") as file_out:
        file_handler = RichHandler(
            show_time=True,
            rich_tracebacks=True,
            tracebacks_show_locals=True,
            level=logging.DEBUG,
            console=Console(file=file_out, width=500),
        )
        logger.addHandler(file_handler)
        async with asyncio.TaskGroup() as tg:

            def write_jsonl(result: object) -> None:
                def dump():
                    with jsonl_path.open("a", encoding="utf8") as file_io:
                        _ = file_io.write(f"{result!s}\n")

                _ = tg.create_task(asyncio.to_thread(dump))

            yield write_jsonl



import logging
from pathlib import Path
from typing import Annotated

from cyclopts import App, Parameter
from rich.traceback import install

from bunkr import __package_name__, __version__
from bunkr.api import BunkrAPI
from bunkr.config import Config
from bunkr.logger import CONSOLE, setup_logger
from bunkr.uploader import BunkrUploader

logger = logging.getLogger(__name__)


install(width=200)


async def _upload(path: Path, recurse: bool, config: Config) -> None:
    async with setup_logger() as json_logger:
        logger.debug(f'Uploading "{path}"')
        logger.debug(f"Using params: \n {config.model_dump_json(indent=4)}")
        async with BunkrUploader(config, upload_callback=json_logger) as client:
            results = await client.upload(path, recurse=recurse)
            for result in results:
                info = f"success: {result.result.success}, url: {result.result.files[0].url}"
                logger.info(f"{result.file.name}: {info}")


app = App(
    name=__package_name__,
    help="Upload files to bunkr.cr",
    version=__version__,
    default_parameter=Parameter(negative_iterable=[]),
)


@app.command()
async def upload(
    path: Annotated[Path, Parameter(help="File or directory to look for files in to upload")],
    /,
    *,
    recurse: Annotated[
        bool, Parameter(help="Read files in PATH recursely", negative_bool=[])
    ] = False,
    config: Annotated[Config, Parameter(name="*")],
) -> None:
    """Upload a file or files to bunkr"""
    await _upload(path, recurse, config)


@app.command()
async def album(url_or_slug: str, /) -> None:
    """Get all files of a public album as a json response"""
    async with BunkrAPI() as api:
        album = await api.public_album(url_or_slug)
        CONSOLE.print(str(album))


if __name__ == "__main__":
    app()



"""Async versions of builtins and some path operations"""

from __future__ import annotations

import asyncio
import dataclasses
from collections.abc import Awaitable
from typing import IO, TYPE_CHECKING, Any, AnyStr, Generic, Self, overload

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Awaitable, Iterable
    from pathlib import Path

    from _typeshed import OpenBinaryMode, OpenTextMode


@dataclasses.dataclass(slots=True, eq=False)
class AsyncIO(Generic[AnyStr]):
    """An asynchronous context manager wrapper for a file object."""

    _coro: Awaitable[IO[AnyStr]]
    _io: IO[AnyStr] = dataclasses.field(init=False)

    async def __aenter__(self) -> Self:
        self._io = await self._coro
        return self

    async def __aexit__(self, *_) -> None:
        return await asyncio.to_thread(self._io.close)

    async def __aiter__(self) -> AsyncIterator[AnyStr]:
        while True:
            line = await self.readline()
            if line:
                yield line
            else:
                break

    async def read(self, size: int = -1) -> AnyStr:
        return await asyncio.to_thread(self._io.read, size)

    async def readline(self) -> AnyStr:
        return await asyncio.to_thread(self._io.readline)

    async def readlines(self) -> list[AnyStr]:
        return await asyncio.to_thread(self._io.readlines)

    async def write(self, b: AnyStr, /) -> int:
        return await asyncio.to_thread(self._io.write, b)

    async def writelines(self, lines: Iterable[AnyStr], /) -> None:
        return await asyncio.to_thread(self._io.writelines, lines)


@overload
def open(
    path: Path,
    mode: OpenBinaryMode,
    buffering: int = ...,
    encoding: str | None = ...,
    errors: str | None = ...,
    newline: str | None = ...,
) -> AsyncIO[bytes]: ...


@overload
def open(
    path: Path,
    mode: OpenTextMode = ...,
    buffering: int = ...,
    encoding: str | None = ...,
    errors: str | None = ...,
    newline: str | None = ...,
) -> AsyncIO[str]: ...


def open(
    path: Path,
    mode: str = "r",
    buffering: int = -1,
    encoding: str | None = None,
    errors: str | None = None,
    newline: str | None = None,
) -> AsyncIO[Any]:
    coro = asyncio.to_thread(path.open, mode, buffering, encoding, errors, newline)
    return AsyncIO(coro)
